==> mars-0.9.2.tar.gz: mars-0.9.2/src/host/lib/mars_task.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified
 *  version of the Library in object code or as an executable provided
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete
 *   corresponding machine-readable source code for the modified version
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer
 *   for a complete machine-readable copy of the corresponding source
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <elf.h>
#include <string.h>
#include <malloc.h>

#include "mars/mars_task.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int mars_task_initialize(struct mars_context *mars,
			struct mars_task_id *id,
			struct mars_task_params *params)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(params, MARS_ERROR_NULL);
	MARS_CHECK_RET(params->elf_image, MARS_ERROR_PARAMS);

	int ret;
	struct mars_task_context *task;
	Elf32_Ehdr *ehdr = (Elf32_Ehdr *)params->elf_image;
	Elf32_Phdr *phdr = (Elf32_Phdr *)((void *)ehdr + ehdr->e_phoff);

	//MARS_CHECK_RET(ehdr->e_phnum == 1, MARS_ERROR_FORMAT);
	MARS_CHECK_RET(phdr->p_type == PT_LOAD, MARS_ERROR_FORMAT);
	MARS_CHECK_RET(phdr->p_vaddr == MARS_TASK_ELF_VADDR,
			MARS_ERROR_FORMAT);

	/* initialize task context */
	id->mars_context_ea = (uint64_t)(uintptr_t)mars;

	/* copy the task name into task id */
	if (params->name) {
		MARS_CHECK_RET(strlen(params->name) < MARS_TASK_NAME_LEN_MAX,
				MARS_ERROR_PARAMS);
		strcpy((char *)&id->name, params->name);
	}

	/* begin process to add the task to the workload queue */
	ret = workload_queue_add_begin(mars->workload_queue, &id->workload_id,
				MARS_WORKLOAD_TYPE_TASK,
				(struct mars_workload_context **)&task);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* initialize task specific id */
	memcpy(&task->id, id, sizeof(struct mars_task_id));

	/* initialize the elf parameters */
	task->exec = (uint64_t)(uintptr_t)((void *)ehdr + phdr->p_offset);
	task->exec_size = phdr->p_filesz;
	task->bss_size = phdr->p_memsz - phdr->p_filesz;
	task->vaddr = phdr->p_vaddr;
	task->entry = ehdr->e_entry;

	/* allocate the task context area if specified */
	if (params->context_save_size) {
		MARS_CHECK_CLEANUP_RET(params->context_save_size <=
			MARS_TASK_CONTEXT_SAVE_SIZE_MAX,
			workload_queue_add_cancel(mars->workload_queue,
				id->workload_id),
			MARS_ERROR_PARAMS);

		task->context_save_size =
			params->context_save_size;
		task->context_save_area = (uint64_t)(uintptr_t)
			memalign(MARS_TASK_CONTEXT_SAVE_ALIGN,
				task->context_save_size);

		MARS_CHECK_CLEANUP_RET(task->context_save_area,
			workload_queue_add_cancel(mars->workload_queue,
				id->workload_id),
			MARS_ERROR_MEMORY);
	} else {
		task->context_save_size = 0;
		task->context_save_area = 0;
	}

	/* end process to add the task to the workload queue */
	ret = workload_queue_add_end(mars->workload_queue, id->workload_id);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS,
			workload_queue_add_cancel(mars->workload_queue,
				id->workload_id),
			ret);

	MARS_PRINT("Initialize Task Context %d\n", task->id.workload_id);
	MARS_PRINT_TASK_CONTEXT(&task);

	return MARS_SUCCESS;
}

int mars_task_finalize(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_task_context *task;
	struct mars_context *mars = (void *)(uintptr_t)id->mars_context_ea;

	/* begin process to remove the task from the workload queue */
	ret = workload_queue_remove_begin(mars->workload_queue, id->workload_id,
				(struct mars_workload_context **)&task);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* free the allocated task context area if it has one */
	if (task->context_save_size)
		free((void *)(uintptr_t)task->context_save_area);

	/* end process to remove the task from the workload queue */
	ret = workload_queue_remove_end(mars->workload_queue, id->workload_id);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	MARS_PRINT("Finalize Task Context %d\n", id->workload_id);

	return MARS_SUCCESS;
}

int mars_task_schedule(struct mars_task_id *id,
			struct mars_task_args *args,
			uint8_t priority)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_task_context *task;
	struct mars_context *mars = (void *)(uintptr_t)id->mars_context_ea;

	/* begin process to schedule the workload in the workload queue */
	ret = workload_queue_schedule_begin(mars->workload_queue,
					id->workload_id, priority,
					(struct mars_workload_context **)&task);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* initialize task specific context variables */
	task->stack = 0;
	if (args)
		memcpy(&task->args, args, sizeof(struct mars_task_args));

	/* end process to schedule the workload in the workload queue */
	ret = workload_queue_schedule_end(mars->workload_queue,
					id->workload_id);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	MARS_PRINT("Schedule Task Context %d\n", id->workload_id);

	return MARS_SUCCESS;
}

int mars_task_wait(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_task_context *task;
	struct mars_context *mars = (void *)(uintptr_t)id->mars_context_ea;

	/* blocking wait for workload completion */
	ret = workload_queue_wait(mars->workload_queue, id->workload_id,
				(struct mars_workload_context **)&task);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	return MARS_SUCCESS;
}

int mars_task_try_wait(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_task_context *task;
	struct mars_context *mars = (void *)(uintptr_t)id->mars_context_ea;

	/* non-blocking wait for workload completion */
	ret = workload_queue_try_wait(mars->workload_queue, id->workload_id,
				(struct mars_workload_context **)&task);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	return MARS_SUCCESS;
}

==> mars-08.06.26.tar.gz: mars-08.06.26/src/host/lib/mars_task.c <==
/*
 * Copyright 2008 Sony Corporation of America
 *
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this Library and associated documentation files (the
 * "Library"), to deal in the Library without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Library, and to
 * permit persons to whom the Library is furnished to do so, subject to
 * the following conditions:
 *
 *  The above copyright notice and this permission notice shall be 
 *  included in all copies or substantial portions of the Library.
 *
 *  If you modify the Library, you may copy and distribute your modified 
 *  version of the Library in object code or as an executable provided 
 *  that you also do one of the following:
 *
 *   Accompany the modified version of the Library with the complete 
 *   corresponding machine-readable source code for the modified version 
 *   of the Library; or,
 *
 *   Accompany the modified version of the Library with a written offer 
 *   for a complete machine-readable copy of the corresponding source 
 *   code of the modified version of the Library.
 *
 *
 * THE LIBRARY IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * LIBRARY OR THE USE OR OTHER DEALINGS IN THE LIBRARY.
 */

#include <string.h>
#include <malloc.h>

#include "mars/mars_task.h"
#include "mars/mars_workload_queue.h"
#include "mars/mars.h"
#include "mars/mars_elf.h"
#include "mars/mars_error.h"
#include "mars/mars_debug.h"

int mars_task_initialize(struct mars_context *mars,
			struct mars_task_id *id,
			struct mars_task_params *params)
{
	MARS_CHECK_RET(mars, MARS_ERROR_NULL);
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(params, MARS_ERROR_NULL);

	int ret;
	struct mars_workload_context workload;
	struct mars_task_context *task = &workload.type_context.task;

	/* initialize task context */
	task->id.mars_context_ea = (uint64_t)(uintptr_t)mars;

	/* parse the elf parameter */
	ret = mars_elf_parse(params->elf_image, &task->exec, &task->exec_size,
				&task->bss_size, &task->vaddr, &task->entry);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* copy the task name into task id */
	if (params->name) {
		MARS_CHECK_RET(strlen(params->name) < MARS_TASK_NAME_LEN_MAX,
				MARS_ERROR_PARAMS);
		strcpy((char *)&task->id.name, params->name);
	}

	/* allocate the task context area if specified */
	if (params->context_save_size) {
		MARS_CHECK_RET(params->context_save_size <=
			MARS_TASK_CONTEXT_SAVE_SIZE_MAX, MARS_ERROR_PARAMS);

		task->context_save_size =
			params->context_save_size;
		task->context_save_area = (uint64_t)(uintptr_t)
			memalign(MARS_TASK_CONTEXT_SAVE_ALIGN,
				task->context_save_size);

		MARS_CHECK_RET(task->context_save_area, MARS_ERROR_MEMORY);
	} else {
		task->context_save_size = 0;
		task->context_save_area = 0;
	}

	mars_mutex_lock(mars->mutex);

	/* set the workload context type to type task */
	workload.type = MARS_WORKLOAD_TYPE_TASK;

	/* add the workload to the workload queue */
	ret = workload_queue_add(mars->workload_queue, &task->id.workload_id,
				&workload);
	MARS_CHECK_CLEANUP_RET(ret == MARS_SUCCESS,
				free((void *)task->context_save_area), ret);
	MARS_PRINT_WORKLOAD_CONTEXT(&workload);

	/* copy the task id into return id */
	*id = task->id;

	mars_mutex_unlock(mars->mutex);

	MARS_PRINT("Initialize Task Context %d\n", task->id.workload_id);

	return MARS_SUCCESS;
}

int mars_task_finalize(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_context *mars = (struct mars_context *)id->mars_context_ea;
	struct mars_workload_context workload;
	struct mars_task_context *task = &workload.type_context.task;

	mars_mutex_lock(mars->mutex);

	/* remove the workload from the workload queue */
	ret = workload_queue_remove(mars->workload_queue, id->workload_id,
				&workload);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* free the allocated task context area if it has one */
	if (task->context_save_size)
		free((void *)task->context_save_area);

	mars_mutex_unlock(mars->mutex);

	MARS_PRINT("Finalize Task Context %d\n", task->id.workload_id);

	return MARS_SUCCESS;
}

int mars_task_schedule(struct mars_task_id *id,
			struct mars_task_args *args,
			uint8_t priority)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_context *mars = (struct mars_context *)id->mars_context_ea;
	struct mars_workload_context *workload = NULL;
	struct mars_task_context *task = NULL;

	mars_mutex_lock(mars->mutex);

	/* get workload context pointer from the workload queue */
	ret = workload_queue_context(mars->workload_queue, id->workload_id,
				&workload);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	/* initialize task specific context variables */
	task = &workload->type_context.task;
	task->stack = 0;
	if (args)
		memcpy(&task->args, args, sizeof(struct mars_task_args));

	/* schedule the workload in the workload queue */
	ret = workload_queue_schedule(mars->workload_queue, id->workload_id,
				priority);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	mars_mutex_unlock(mars->mutex);

	MARS_PRINT("Schedule Task Context %d\n", task->id.workload_id);

	return MARS_SUCCESS;
}

int mars_task_wait(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_context *mars = (struct mars_context *)id->mars_context_ea;

	ret = workload_queue_wait(mars->workload_queue, id->workload_id);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	return MARS_SUCCESS;
}

int mars_task_try_wait(struct mars_task_id *id)
{
	MARS_CHECK_RET(id, MARS_ERROR_NULL);
	MARS_CHECK_RET(id->mars_context_ea, MARS_ERROR_PARAMS);
	MARS_CHECK_RET(id->workload_id < MARS_WORKLOAD_MAX, MARS_ERROR_PARAMS);

	int ret;
	struct mars_context *mars = (struct mars_context *)id->mars_context_ea;

	ret = workload_queue_try_wait(mars->workload_queue, id->workload_id);
	MARS_CHECK_RET(ret == MARS_SUCCESS, ret);

	return MARS_SUCCESS;
}
